/*
 * Copyright (C) 2012 Chris Burdess <dog@gnu.org>
 *
 * Gnome Shell Biff is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Gnome Shell Biff is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Gnome Documents; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

const Gio = imports.gi.Gio;
const Lang = imports.lang;
const Extension = imports.ui.extensionSystem.extensions["biff@bluezoo.org"];
const Sasl = Extension.sasl;

function IMAPConnection(host, port, username, password, ssl, mailbox, recent_callback) {
    this.init(host, port, username, password, ssl, mailbox, recent_callback);
}

IMAPConnection.prototype = {

init : function(host, port, username, password, ssl, mailbox, recent_callback) {
    this.IMAP_AUTH_METHODS = new Array("CRAM-MD5", "PLAIN");
    this.host = host;
    this.port = port;
    this.username = username;
    this.id = username+"@"+host+":"+port;
    this.password = password;
    this.try_starttls = !ssl;
    this.can_starttls = false;
    if (mailbox == "") {
        mailbox = "INBOX";
    }
    this.mailbox = mailbox;
    this.recent_callback = recent_callback;
    this.tag = 0;
    this.auth_capabilities = new Array();
    this.sasl_client = null;
    this.socket = Gio.SocketClient.new();
    if (ssl) {
        this.socket.set_tls(true);
        this.socket.set_tls_validation_flags(0);
    }
    this.socket.set_timeout(10);
},

connect : function() {
    try {
        // TODO aync connect
        this.conn = this.socket.connect_to_host(this.host, this.port, null);
        this.is = Gio.DataInputStream.new(this.conn.get_input_stream());
        this.os = Gio.DataOutputStream.new(this.conn.get_output_stream());
        this.is.set_buffer_size(32768);
        this.readline(this.welcome_callback);
    } catch (err) {
        global.log("Error connecting to "+this.host+": "+err.message);
    }
},

new_tag : function() {
    this.tag++;
    return "A" + this.tag;
},

readline : function(callback) {
    this.is.read_line_async(0, null, Lang.bind(this, function (stream, result) {
        let buf = this.is.read_line_finish(result);
        let line = new String(buf[0]).substr(0, buf[0].length - 1);
        //global.log(this.host+"< "+line);
        callback.apply(this, [line]);
    }));
},

writeline : function(line) {
    //global.log(this.host+"> "+line);
    this.os.put_string(line, null);
},

welcome_callback : function(line) {
    //global.log("welcome_callback: "+line);
    var tokens = this.tokenize(line);
    if (tokens[0] === "*" && tokens[1] === "OK") {
        if (this.is_list(tokens[2])) {
            if (this.handle_capability_list(this.tokenize_list(tokens[2]))) {
                this.starttls();
                return;
            }
        } 
        this.capability();
        return;
    }
    global.log("Unexpected IMAP response from "+this.host+": "+line);
    this.close();
},

capability : function() {
    //global.log("capability");
    this.writeline(this.new_tag()+" CAPABILITY\r\n");
    this.readline(this.capability_callback);
},

capability_callback : function(line) {
    //global.log("capability_callback: "+line);
    var tokens = this.tokenize(line);
    if (tokens[0] === "*") {
        tokens.shift();
        this.handle_capability_list(tokens);
        this.readline(this.capability_callback);
    } else if (tokens[1] === "OK") {
        this.starttls();
    } else {
        global.log("Unexpected SELECT response from "+this.host+": "+line);
        this.close();
    }
},

handle_capability_list : function(list) {
    //global.log("capability="+list+" try_starttls="+this.try_starttls);
    if (list[0] === "CAPABILITY") {
        for (var i = 1; i < list.length; i++) {
            if (list[i] === "STARTTLS" && this.try_starttls) {
                this.can_starttls = true;
            } else if (list[i].substr(0, 5) === "AUTH=") {
                this.auth_capabilities.push(list[i].substr(5));
            }
        }
        return true;
    }
    return false;;
},

starttls : function() {
    if (this.try_starttls && this.can_starttls) {
        this.writeline(this.new_tag()+" STARTTLS\r\n");
        this.readline(this.starttls_callback);
    } else {
        this.authenticate();
    }
},

starttls_callback : function(line) {
    //global.log("starttls_callback: "+line);
    var tokens = this.tokenize(line);
    if (tokens[1] === "OK") {
        var addr = Gio.InetAddress.new_any(2); // IPv4
        var iaddr = Gio.InetSocketAddress.new(addr, this.port);
        try {
            var sconn = Gio.tls_client_connection_new(this.conn, iaddr);
            sconn.set_validation_flags(0);
            sconn.handshake(null);
            this.conn = sconn;
            this.is = Gio.DataInputStream.new(this.conn.get_input_stream());
            this.os = Gio.DataOutputStream.new(this.conn.get_output_stream());
            this.try_starttls = false;
        } catch (err) {
            global.log("Error starting TLS on "+this.host+": "+err.message);
        }
        this.authenticate();
    } else {
        global.log("Host "+this.host+" advertised STARTTLS but failed to acknowledge: "+line);
        close();
    }
},

get_sasl_client : function() {
    for (var i = 0; i < this.IMAP_AUTH_METHODS.length; i++) {
        var method = this.IMAP_AUTH_METHODS[i];
        if (this.auth_capabilities.indexOf(method) != -1) {
            if (method === "CRAM-MD5") {
                this.sasl_client = new sasl.CramMD5(this.username, this.password);
            } else if (method == "PLAIN") {
                this.sasl_client = new sasl.Plain(this.username, this.password);
            }
        }
    }
},

authenticate : function() {
    //global.log("authenticate");
    this.get_sasl_client();
    if (this.try_starttls && (this.sasl_client == null || this.sasl_client.is_plaintext())) {
        // this is over an unencrypted connection,
        // we should be using a secure authentication method at least
        global.log("No secure authentication method for "+this.username+" on "+this.host);
        this.logout();
        return;
    }
    if (this.sasl_client != null) {
        //global.log("Authenticating to "+this.host+" using "+this.sasl_client.get_name());
        let cmd = this.new_tag()+" AUTHENTICATE "+this.sasl_client.get_name();
        if (this.sasl_client.has_initial_response()) {
            let response = this.sasl_client.evaluate_challenge(null);
            let encoded = GLib.base64.encode(response, response.length);
            cmd = cmd + " " + encoded;
        }
        this.writeline(cmd+"\r\n");
    } else {
        // fall back to LOGIN
        //global.log("Login to "+this.host);
        this.writeline(this.new_tag()+" LOGIN "+this.quote(this.username)+" "+this.quote(this.password)+"\r\n");
    }
    this.readline(this.auth_callback);
},

auth_callback : function(line) {
    //global.log("auth_callback: "+line);
    var tokens = this.tokenize(line);
    if (tokens[0] === "*") {
        if (this.sasl_client != null) {
            let [challenge, clen] = GLib.base64_decode(line.substring(2).trim());
            let response = this.sasl_client.evaluate_challenge(challenge);
            let encoded = GLib.base64.encode(response, response.length);
            this.writeline(encoded + "\r\n");
        }
        this.readline(this.auth_callback);
    } else if (tokens[1] === "OK") {
        this.stat();
    } else {
        global.log("Authentication failed for "+this.username+" on "+this.host);
        this.close();
    }
},

stat : function() {
    this.writeline(this.new_tag()+" STATUS "+this.mailbox+" (unseen)\r\n");
    this.readline(this.stat_callback);
},

stat_callback : function(line) {
    //global.log("stat_callback: "+line);
    var tokens = this.tokenize(line);
    if (tokens[0] === "*") {
        if (tokens[1] === "STATUS") {
            line = new String(line);
            var lp = line.lastIndexOf(")");
            var sp = line.lastIndexOf("(UNSEEN ", lp);
            if (sp > 0 && lp > sp) {
                var unseen = line.substring(sp + 8, lp);
                this.recent_callback.apply(this, [this.id, parseInt(unseen)]);
            }
        }
        this.readline(this.stat_callback);
    } else if (tokens[1] === "OK") {
        this.logout();
    } else {
        global.log("Unexpected SELECT response from "+this.host+": "+line);
        this.close();
    }
},

logout: function() {
    this.writeline(this.new_tag()+" LOGOUT\r\n");
    this.close();
},

close : function() {
    //global.log("Closing connection to "+this.host);
    try {
        this.conn.close(null);
    } catch (err) {
        global.log("Error closing connection to "+this.host+": "+err.message);
    }
},

quote : function(text) {
    if (text.length == 0 || text.indexOf(" ") != -1 || text.indexOf("%") != -1) {
        return "\"" + text + "\"";
    }
    return text;
},

tokenize : function(line) {
    line = new String(line);
    var tokens = new Array();
    var start = 0;
    var end;
    var last = ' ';
    var depth = 0;
    for (end = start; end < line.length; end++) {
        var c = line.charAt(end);
        if (c === '\r' || c === '\n') {
            break;
        }
        if (c === '[') {
            depth++;
        } else if (c === ']') {
            depth--;
        } else if (c === ' ' && depth == 0) {
            if (last != ' ') {
                tokens.push(line.substring(start, end));
            }
            start = end + 1;
        }
        last = c;
    }
    if (end > start) {
        tokens.push(line.substring(start, end));
    }
    return tokens;
},

is_list : function(token) {
    return token.charAt(0) === '[' && token.charAt(token.length - 1) === ']';
},

tokenize_list : function(token) {
    return this.tokenize(token.substring(1, token.length - 1));
}

};
